/* 
 * Copyright (C) 2008 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#include "pbmreader.h"
#include "util.h"
#include "bittwiddler.h"

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/* WARNING: pbmreader doesn't handle files with crazy comments yet */
/* WARNING2: pbmreader doesn't handle multiple images in a single file */
bool_t create_pbmreader(pbmreader_t *p, int fd) {
  if( p ) {
    
    p->type = tUNDEF;
    p->width = 0;
    p->height = 0;
    /* duplicate the file descriptor, because gzclose will destroy it */
    p->f = gzdopen(dup(fd), "rb");
    if( p->f ) {
      p->bufsize = 4096;
      /* buffer has one extra byte for a NULL marker */
      p->buf = malloc((p->bufsize + 1) * sizeof(byte_t));
      if( p->buf ) {
	p->buf[p->bufsize] = '\0'; /* for string functions */
	p->bufpos = p->buf; /* read position */
	p->bufend = p->buf; /* end of data */
      } else {
	p->bufsize = 0;
      }
      if( !get_header_info_pbmreader(p) ) {
	free_pbmreader(p);
	return FALSE;
      }
      return TRUE;
    }
  }
  return FALSE;
}

bool_t read_integer_pbmreader(pbmreader_t *p, unsigned int *retval) {
  char *q;
  if( p && retval && (p->bufend > p->bufpos) ) {
    /* this uses the NULL at the end of the buffer */
    *retval = (unsigned int)strtoul((const char *)p->bufpos, &q, 10); 
    if( ((byte_t *)q > p->bufpos) && ((byte_t *)q <= p->bufend) ) {
      p->bufpos = (byte_t *)q;
      return TRUE;
    }
  }
  return FALSE;
}

/* convert P1 to P4, and return number of packed bits */
/* rowcounter is needed for padding the last byte of each row */ 
int packbits_pbmreader(pbmreader_t *p, unsigned int *rowcounter) {
  bittwiddler_t bt;
  int bitsread;
  unsigned int pixel, rc;
  byte_t *writepos;

  if( p && p->buf && (p->type == tP1) && rowcounter) {
    bitsread = 0;
    writepos = p->buf;
    setbyte_bittwiddler(&bt, 0);
    rc = *rowcounter;
    while( p->bufpos < p->bufend ) {
      switch(p->bufpos[0]) {
      case '0': 
	pixel = 0; 
	p->bufpos++;
	break;
      case '1': 
	pixel = 1; 
	p->bufpos++;
	break;
      case ' ':
      case '\n':  /* ignore this */
	p->bufpos++; 
	continue;
      default:
	return -1; /* bad */
      }
      bitsread++;

      if( !appendbit_bittwiddler(&bt, pixel) ) {
	*writepos++ = getbyte_bittwiddler(&bt);
	setbyte_bittwiddler(&bt, 0);
	appendbit_bittwiddler(&bt, pixel);
      }

      /* each row is padded to a byte boundary, then we exit */
      if( ++rc >= p->width ) {
	rc = 0;
	*writepos++ = getbyte_bittwiddler(&bt);
	setbyte_bittwiddler(&bt, 0);
	goto done;
      }
    }
    /* flush last byte */
    if( numbits_bittwiddler(&bt) > 0 ) {
      *writepos++ = getbyte_bittwiddler(&bt);
      setbyte_bittwiddler(&bt, 0);
    }
  done:
    *rowcounter = rc;
    return bitsread;
  }
  return -1;
}

/* in P4 format, each row is padded to a byte boundary */
int getbits_pbmreader(pbmreader_t *p, unsigned int *rowcounter) {
  unsigned int rc;
  int n, bitsread;
  if( p && p->buf && (p->type == tP4) && rowcounter) {
    rc = *rowcounter;
    bitsread = (p->bufend - p->bufpos) * BITSPERBYTE;
    bitsread = MIN(bitsread, (p->width - rc));
    rc += bitsread;
    rc = (rc >= p->width) ? 0 : rc;

    n = bits2bytes(bitsread);
    memmove(p->buf, p->bufpos, n); /* inefficient, but similar to packbits */
    p->bufpos += n;

    *rowcounter = rc;
    return bitsread;
  }
  return -1;
}


bool_t get_header_info_pbmreader(pbmreader_t *p) {
  if( p &&
      read_first_buffer_pbmreader(p) ) {
    if( p->bufend - p->bufpos < 3 ) {
      return FALSE;
    }
    if( memcmp(p->buf, "P1\n",3) == 0 ) {
      p->type = tP1;
    } else if( memcmp(p->buf, "P4\n",3) == 0 ) {
      p->type = tP4;
    } else {
      return FALSE;
    }
    p->bufpos += 3;

    /* this handles very simple (and short) comments only */
    if( *p->bufpos == '#' ) {
      while( (p->bufpos < p->bufend) && (*p->bufpos != '\n') ) { 
	p->bufpos++; 
      }
    }
    if( read_integer_pbmreader(p, &p->width) &&
	read_integer_pbmreader(p, &p->height) ) {
      if( *p->bufpos == '\n') { 
	p->bufpos++; 
      }
      return TRUE;
    }
  }
  return FALSE;
}

bool_t free_pbmreader(pbmreader_t *p) {
  if( p ) {
    if( p->f ) {
      gzclose(p->f);
      p->f = NULL;
    }
    if( p->buf ) {
      free(p->buf);
      p->buf = NULL;
    }
  }
  return FALSE;
}

bool_t read_first_buffer_pbmreader(pbmreader_t *p) {
  if( p && p->f ) {
    gzseek(p->f, 0, SEEK_SET);
    return read_next_buffer_pbmreader(p);
  } 
  return FALSE;
}

bool_t read_next_buffer_pbmreader(pbmreader_t *p) {
  int n;
  if( p && p->f && p->buf ) {
    n = gzread(p->f, p->buf, p->bufsize);
    if( n == -1 ) {
      errormsg(E_WARNING, "file read error\n");
      free_pbmreader(p);
    } else {
      p->bufpos = p->buf;
      p->bufend = p->buf + n;
    }
    return (n>0);
  }
  return FALSE;
}

/* void print_bt(const byte_t b) { */
/*   unsigned int v; */
/*   bittwiddler_t bt; */
/*   printf("\ncheckbyte "); */
/*   setbyte_bittwiddler(&bt, b); */
/*   //  readfirstbit_bittwiddler(&bt, &v); */
/*   rewind_bittwiddler(&bt); */
/*   //    printf("%d", v); */
/*   while(readnextbit_bittwiddler(&bt, &v)) { */
/*     printf("%d", v); */
/*   } */
/*   printf("\n"); */
/* } */


bool_t fillimage_pixl0_pbmreader(pbmreader_t *p, imagebuffer_t *ib) {
  int bits;
  unsigned int rowcounter;
  if( p && p->f && p->buf && (p->bufsize > 0) && ib &&
      (p->width > 0) && (p->height > 0) ) {
    /* this code assumes that we have read the header, so that
       we are positioned in the data section */
    seekpos_imagebuffer(ib, 0, 0);
    rowcounter = 0;
    do {
      do {
	switch(p->type) {
	case tP1:
	  bits = packbits_pbmreader(p, &rowcounter);
	  partfill_bw_imagebuffer(ib, p->buf, bits, p->width, p->height, 0);
	  break;
	case tP4:
	  bits = getbits_pbmreader(p, &rowcounter);
	  partfill_bw_imagebuffer(ib, p->buf, bits, p->width, p->height, 0); 
	  break;
	default:
	  return FALSE;
	}
      } while( p->bufpos < p->bufend );
    } while( read_next_buffer_pbmreader(p) );
    
    /* file was automatically closed on error */
    return (p->f != NULL);
  }
  return FALSE;
}
